#include <msp/datafile/parser.h>
#include <msp/datafile/statement.h>
#include <msp/datafile/writer.h>
#include <msp/fs/utils.h>
#include <msp/io/memory.h>
#include <msp/strings/format.h>
#include "packer.h"
#include "tool.h"

using namespace std;
using namespace Msp;

Packer::Packer(DataTool &t):
	tool(t),
	tmp_file(tempfile()),
	dir_alloc(0)
{ }

IO::BufferedFile *Packer::tempfile()
{
	FS::Path tmpdir;
	const char *tmp_env = getenv("TMPDIR");
	if(tmp_env)
		tmpdir = tmp_env;
	else
		tmpdir = "/tmp";

	for(unsigned i=0;; ++i)
	{
		try
		{
			FS::Path filename = tmpdir/format("mspdatatool.%d", i);
			IO::BufferedFile *file = new IO::BufferedFile(filename.str(), IO::M_RDWR, IO::File::C_NEW);
			FS::unlink(filename.str());
			return file;
		}
		catch(const IO::file_already_exists &)
		{
			continue;
		}
	}
}

Packer::~Packer()
{
	delete tmp_file;
}

void Packer::pack_file(const string &fn)
{
	if(!tmp_file)
		throw logic_error("Packer::pack_file");

	IO::SeekOffset offset = tmp_file->tell();

	IO::BufferedFile in(fn);
	ObjectList objects;
	bool raw = detect_raw(in);
	if(raw)
		transfer_raw(in, *tmp_file);
	else
		transfer_datafile(in, fn, *tmp_file, objects);

	unsigned length = tmp_file->tell()-offset;

	DataFile::Statement st("file");
	st.append(FS::basename(fn));
	st.sub.push_back((DataFile::Statement("slice"), offset, length));
	for(list<Object>::const_iterator i=objects.begin(); i!=objects.end(); ++i)
		st.sub.push_back((DataFile::Statement("object"), i->name, i->keyword));
	directory.push_back(st);
	dir_alloc += 100+objects.size()*100;
}

bool Packer::detect_raw(IO::Seekable &in)
{
	IO::SeekOffset offset = in.tell();
	bool raw = false;
	DataFile::Parser parser(in, "-");
	while(!raw && parser && in.tell()<10240)
	{
		try
		{
			parser.parse();
		}
		catch(...)
		{
			raw = true;
		}
	}
	in.seek(offset, IO::S_BEG);
	return raw;
}

void Packer::transfer_datafile(IO::Base &in, const string &fn, IO::Base &out, ObjectList &objects)
{
	DataFile::Parser parser(in, fn);

	DataFile::Writer *writer = tool.create_writer(out);

	bool collection = FS::extpart(fn)==".mdc";
	while(parser)
	{
		DataFile::Statement st = parser.parse(true);
		if(st.valid)
		{
			if(collection && !st.control)
			{
				if(st.get_signature()=="s")
				{
					Object obj;
					obj.name = st.args[0].get<string>();
					obj.keyword = st.keyword;
					objects.push_back(obj);
				}
				else
					collection = false;
			}
			if(!st.control || st.keyword=="__src")
				writer->write(st);
		}
	}
	writer->write(DataFile::Statement("__end"));
	delete writer;
}

void Packer::transfer_raw(IO::Base &in, IO::Base &out)
{
	while(!in.eof())
	{
		char buf[4096];
		unsigned len = in.read(buf, sizeof(buf));
		if(len==0)
			break;
		out.write(buf, len);
	}
}

void Packer::create_pack(const string &fn)
{
	vector<char> dir_buffer(dir_alloc);
	IO::Memory mem(&dir_buffer[0], dir_buffer.size(), IO::M_WRITE);

	unsigned base_offset = 0;
	while(1)
	{
		mem.seek(0, IO::S_BEG);

		DataFile::Writer *writer = tool.create_writer(mem);

		for(list<DataFile::Statement>::const_iterator i=directory.begin(); i!=directory.end(); ++i)
			writer->write(*i);
		if(base_offset==0)
			base_offset = mem.tell();
		writer->write((DataFile::Statement("base_offset"), base_offset));
		writer->write(DataFile::Statement("__end"));
		delete writer;

		unsigned dir_size = mem.tell();
		if(dir_size<=base_offset)
			break;
		base_offset = dir_size;
	}

	IO::File out(fn, IO::M_WRITE);
	out.write(&dir_buffer[0], base_offset);
	tmp_file->seek(0, IO::S_BEG);
	unsigned bufsize = 1048576;
	char *buf = new char[bufsize];
	while(!tmp_file->eof())
	{
		unsigned len = tmp_file->read(buf, bufsize);
		if(!len)
			break;
		out.write(buf, len);
	}
	delete[] buf;
}
